/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.plugin.view;

import org.opengion.fukurou.util.StringUtil;
import org.opengion.fukurou.util.Attributes;

import org.opengion.hayabusa.db.DBColumn;
import org.opengion.hayabusa.db.DBColumnConfig;
import org.opengion.hayabusa.db.DBTableModel;
import org.opengion.hayabusa.resource.ResourceFactory;
import org.opengion.hayabusa.resource.ResourceManager;
import org.opengion.hayabusa.resource.LabelData;
import org.opengion.hayabusa.common.HybsSystemException;

/**
 * 指定の行-列と、動的ｶﾗﾑのﾃｰﾌﾞﾙを作成するクラスです。
 *
 * AbstractViewForm により、setter/getterメソッドのデフォルト実装を提供しています。
 * 各HTMLのタグに必要な setter/getterメソッドのみ，追加定義しています。
 *
 * RENDERER,EDITOR,DBTYPE を使用する代わりに、簡易的な DATA_TYPE で決定します。
 *
 * GG10 履歴ﾃｰﾌﾞﾙ (書き込むﾃｰﾌﾞﾙ)
 *   ﾄｰｸﾝ(TOKEN)			必須ｷｰ(ﾄｰｸﾝ)
 *   更新ｶｳﾝﾀ(UPCNT)		← 未使用(同一ﾄｰｸﾝでの最大値が有効:逆順検索なので、最初に見つかった値を採用)
 *   値ﾃﾞｰﾀ(VAL)			必須ｷｰ(値の設定)
 *   単位(TANI)				ﾌｨｰﾙﾄﾞの後ろに追記(連続の場合は、一番最後のみ)
 *   判定結果(JUDG)			0:未決 1:必須 2:任意 3:不要 4:合格 5:不合格 6:保留 7:警告
 *   判定理由(RIYU)			上限、下限で判定した結果が入っている。titleに入れてﾎﾟｯﾌﾟｱｯﾌﾟさせる
 *
 * GG01 ﾄｰｸﾝﾏｽﾀ (GG02がnullの時)
 *   ﾄｰｸﾝ名称(TKN_NM)		← 未使用(GG02 のﾄｰｸﾝ名称が未設定の場合…SQL文で処理済み)
 *   表示桁数(VIEW_LEN)		ﾃｷｽﾄﾌｨｰﾙﾄﾞの長さｾｯﾄ
 *   ﾃﾞｰﾀ型(DATA_TYPE)		EDITORを決定
 *   ﾄｰｸﾝｸﾞﾙｰﾌﾟ(TKN_GRP)	(未使用=GG03のSEL_KEY の条件に使用している)
 *
 * GG03 選択ﾏｽﾀ (GG01 ﾄｰｸﾝｸﾞﾙｰﾌﾟの名称とﾏｰｶｰ)
 *   選択名称(SEL_NM)		ﾄｰｸﾝｸﾞﾙｰﾌﾟ名で、fieldset のｷｰﾌﾞﾚｲｸに使用
 *   ﾏｰｶｰ(MARKER)			fieldset のstyle属性として使用
 *
 * GG02 雛形設定ﾏｽﾀ
 *   ﾄｰｸﾝ名称(TKN_NM)		(名称優先)
 *   初期値(DEF_VAL)		値の設定 する場合の初期値に使用
 *   行番号(ROWNO)			ﾄｰｸﾝの並び順を指定
 *   列番号(COLNO)			← 未使用(ﾄｰｸﾝの並び順のサブ)
 *   行数(ROWSPAN)			← 未使用
 *   列数(COLSPAN)			tableのｶﾗﾑを複数つなげる場合
 *   登録方法(CDREC)		0:未決 1:必須 2:任意 3:不要
 *   表示方法(CDDISP)		1:ﾗﾍﾞﾙｶﾗﾑ 2:ｶﾗﾑ単発 3:ｶﾗﾑ連続 4:ﾗﾍﾞﾙのみ
 *   異常下限(E_MIN)		異常値の下限判定をﾌｨｰﾙﾄﾞの横に記述
 *   警告下限(W_MIN)		警告値の下限判定をﾌｨｰﾙﾄﾞの横に記述
 *   警告上限(W_MAX)		警告値の上限判定をﾌｨｰﾙﾄﾞの横に記述
 *   異常上限(E_MAX)		異常値の上限判定をﾌｨｰﾙﾄﾞの横に記述
 *
 * @og.group 画面表示
 *
 * @version  7.3
 * @author   Kazuhiko Hasegawa
 * @since    JDK11.0,
 */
public class ViewForm_HTMLTokenTable extends ViewForm_HTMLTable {
	/** このプログラムのVERSION文字列を設定します。	{@value} */
	private static final String VERSION = "7.3.2.1 (2021/03/25)" ;

	private int noToken		= -1;
//	private int noUpcnt		= -1;
	private int noVal		= -1;
	private int noTani		= -1;
	private int noJudg		= -1;
	private int noRiyu		= -1;

	private int noViewlen	= -1;
	private int noType		= -1;

	private int noSelnm		= -1;
	private int noMarker	= -1;

	private int noName		= -1;
	private int noDefval	= -1;
	private int noRowno		= -1;
//	private int noColno		= -1;
//	private int noRowspan	= -1;
	private int noColspan	= -1;

	private int noCdrec		= -1;
	private int noCddisp	= -1;
	private int noEmin		= -1;
	private int noWmin		= -1;
	private int noWmax		= -1;
	private int noEmax		= -1;

	private DBTableModel table ;

	/**
	 * デフォルトコンストラクター
	 *
	 * @og.rev 7.3.2.1 (2021/03/25) 動的ｶﾗﾑのﾃｰﾌﾞﾙを作成する
	 */
	public ViewForm_HTMLTokenTable() { super(); }		// これも、自動的に呼ばれるが、空のメソッドを作成すると警告されるので、明示的にしておきます。

	/**
	 * 初期化します。
	 * ここでは、内部で使用されているキャッシュをクリアし、
	 * 新しいモデル(DBTableModel)と言語(lang) を元に内部データを再構築します。
	 * ただし、設定情報は、以前の状態がそのままキープされています。
	 *
	 * @og.rev 7.3.2.1 (2021/03/25) 動的ｶﾗﾑのﾃｰﾌﾞﾙを作成する
	 *
	 * @param	table	DBTableModelオブジェクト
	 */
	@Override
	public void init( final DBTableModel table ) {
		super.init( table );
		this.table = table;

		noToken		= table.getColumnNo( "TOKEN"			);	// ﾄｰｸﾝ
	//	noUpcnt		= table.getColumnNo( "UPCNT"			);	// 更新ｶｳﾝﾀ
		noVal		= table.getColumnNo( "VAL"				);	// 値
		noTani		= table.getColumnNo( "TANI"		, false );	//
		noJudg		= table.getColumnNo( "JUDG"		, false );	//
		noRiyu		= table.getColumnNo( "RIYU"		, false );	//
		noViewlen	= table.getColumnNo( "VIEW_LEN"	, false );	//
		noType		= table.getColumnNo( "DATA_TYPE", false );	//
		noSelnm		= table.getColumnNo( "SEL_NM"	, false );	//
		noMarker	= table.getColumnNo( "MARKER"	, false );	//
		noName		= table.getColumnNo( "TKN_NM"	, false );	//
		noDefval	= table.getColumnNo( "DEF_VAL"	, false );	//
		noRowno		= table.getColumnNo( "ROWNO"	, false );	//
	//	noColno		= table.getColumnNo( "COLNO"	, false );	//
	//	noRowspan	= table.getColumnNo( "ROWSPAN"	, false );	//
		noColspan	= table.getColumnNo( "COLSPAN" 	, false );	//
		noCdrec		= table.getColumnNo( "CDREC"	, false );	//
		noCddisp	= table.getColumnNo( "CDDISP"	, false );	//
		noEmin		= table.getColumnNo( "E_MIN"	, false );	//
		noWmin		= table.getColumnNo( "W_MIN"	, false );	//
		noWmax		= table.getColumnNo( "W_MAX"	, false );	//
		noEmax		= table.getColumnNo( "E_MAX"	, false );	//
	}

	/**
	 * DBTableModel から HTML文字列を作成して返します。
	 * startNo(表示開始位置)から、pageSize(表示件数)までのView文字列を作成します。
	 * 表示残りデータが pageSize 以下の場合は,残りのデータをすべて出力します。
	 *
	 * @og.rev 7.3.2.1 (2021/03/25) 動的ｶﾗﾑのﾃｰﾌﾞﾙを作成する
	 *
	 * @param  startNo    表示開始位置
	 * @param  pageSize   表示件数
	 *
	 * @return  DBTableModelから作成された HTML文字列
	 * @og.rtnNotNull
	 */
	@Override
	public String create( final int startNo, final int pageSize )  {
		if( getRowCount() == 0 ) { return ""; }	// 暫定処置

//		final int lastNo = getLastNo( startNo, pageSize );
int lastNo = getLastNo( startNo, pageSize );

		final StringBuilder out = new StringBuilder( BUFFER_LARGE )
			.append( getCountForm( startNo,pageSize ) )
			.append( getHeader() );

		final ResourceManager resource = ResourceFactory.newInstance( "ja" ) ;

		int lastRow  = -1;
		int bgClrCnt = 0;
		boolean lastTd = false;			// 最後に </td> が 必要な場合は、true にする。

		// td無しでカラムのみ追加する場合は、最後の</td>の直前にまとめて登録する。
		final StringBuilder valBuf = new StringBuilder( BUFFER_MIDDLE );
		String tani    = "" ;			// 単位は、ループの最後の値を使用するため。
		String bkSelNm = "";			// 最後にセットされた SEL_NM (ｷｰﾌﾞﾚｲｸ用)
		String bkToken = "";			// 更新ｶｳﾝﾀの最大値(逆順検索しているので、最初に見つかったﾄｰｸﾝ)のみ採用する。

		final int clmCnt = getColumnCount();
		for( int row=startNo; row<lastNo; row++ ) {
			final int rowno = rowData( row,noRowno,row-startNo );	// DBに指定した列の番号

			final String token = table.getValue( row,noToken );
	//		final String upcnt = table.getValue( row,noUpcnt );		// 更新ｶｳﾝﾀは、同一ﾄｰｸﾝ内で、一番最初に現れたﾚｺｰﾄﾞのみ使用する。
			if( bkToken.equals( token ) ) {				// 同一ﾄｰｸﾝなら、過去ﾃﾞｰﾀなので取り直し
				table.removeValue( row );	// 物理削除
				row--;
				lastNo--;
				continue;
			}

			bkToken = token;

			final String selNm  = rowData( row,noSelnm ,"" );		// 選択名称(SEL_NM)
			if( lastRow != rowno || !bkSelNm.equals( selNm ) ) {
				if( lastRow >= 0 ) {	// 最初のループだけは、実行しない。
					if( !bkSelNm.equals( selNm ) && !bkSelNm.isEmpty() ) {
						if( lastTd ) {										// 前回、</td> を付けていないため
							valBufOut( row-1,out,valBuf,tani );				// out に、valBuf と tani と、警告異常範囲と終端 td を付ける。
							lastTd=false;
						}
						out.append( "</tr></tbody></table></fieldset>" );
					}
					else {
						if( lastTd ) {		// 前回、</td> を付けていないため
							valBufOut( row-1,out,valBuf,tani );				// out に、valBuf と tani と、警告異常範囲と終端 td を付ける。
							lastTd=false;
						}
						out.append(" </tr>").append( CR );
					}
				}
				lastRow = rowno;

				if( !bkSelNm.equals( selNm ) && !selNm.isEmpty() ) {
					final String style = rowData( row,noMarker," style=\"", "\"" );	// ﾏｰｶｰ(MARKER) は、style属性に追加する。
					out.append( "<fieldset" ).append( style ).append( "><legend>" ).append( selNm )
						.append( "</legend><table><tbody>" )
						.append(" <tr").append( getBgColorCycleClass( bgClrCnt++,row ) ).append('>').append( CR );
				}
				else {
					out.append(" <tr").append( getBgColorCycleClass( bgClrCnt++,row ) ).append('>').append( CR );
				}
				bkSelNm = selNm;
			}

			final String lbl   = rowData( row,noName,token );		// 表示するﾗﾍﾞﾙ(無ければﾄｰｸﾝそのもの)
			final String dbType= rowData( row,noType,"TP_TEXT" );	// DBﾀｲﾌﾟ。未指定時は、TP_TEXT:ﾃｷｽﾄ

			tani = rowData( row,noTani,"(",")" );					// 単位表示(ｶﾗﾑの最後に出すので、ﾙｰﾌﾟの次のﾀｲﾐﾝｸﾞに出す)…最後の単位のみ

			final DBColumn clm = resource.makeDBColumn( dbType,lbl );
			final DBColumnConfig config = clm.getConfig();
			config.setName( "VAL" );								// 登録するｶﾗﾑの名前は、VAL になる。

			if( noViewlen >= 0 ) { config.setFieldSize( table.getValue( row,noViewlen ) ); }	// ﾌｨｰﾙﾄﾞの表示桁数

			final String defVal = rowData( row,noDefval,"" );		// 初期値

			final String val = rowData( row,noVal,defVal );

			final String cdrec = rowData( row,noCdrec ,"0" );		// 0:未決 1:必須 2:任意 3:不要
			final String judg  = rowData( row,noJudg  ,"0" );		// 0:未決 1:必須 2:任意 3:不要 4:合格 5:不合格 6:保留 7:警告
			final Attributes editAttri = config.getEditorAttributes();

			final boolean isEmp = val.isEmpty()													// 本当にempty
									|| ( "TP_CHECK".equals( dbType ) && "0".equals(val) );		// TP_CHECK が '0' は、未設定と同じ

			if( isEmp ) {											// val が未設定の場合は、cdrec をそのままセットする。
				editAttri.add( "class" , "judg" + cdrec );			// 注意：キーは、judg
			}
			else {													// val が設定済みの場合は、judg をセットする。
				editAttri.add( "class" , "judg"  + judg );
			}

			// dbType が、TP_NUMBER の場合は、数値のみ可能とする。
			if( "TP_NUMBER".equals( dbType ) ) {
				editAttri.set( "type" , "number" );
			}

			final String cddisp = rowData( row,noCddisp,"1" );		// 1:ﾗﾍﾞﾙｶﾗﾑ 2:ｶﾗﾑ単発 3:ｶﾗﾑ連続 4:ﾗﾍﾞﾙのみ

			// CDDISP=3:ｶﾗﾑ連続 の場合に、ﾗﾍﾞﾙの末尾が数字の場合は、placeholder にｾｯﾄする。
			if( "3".equals( cddisp ) ) {
				int idx = lbl.length()-1;
				while( idx >= 0 ) {
					final char ch = lbl.charAt( idx );
					if( '0' <= ch && ch <= '9' ) { idx--; }		// 末尾から数字の間、処理を継続
					else { break; }								// 数字でなくなった時点で処理終了
				}
				if( idx >= 0 && idx < lbl.length()-1 ) {		// 末尾に何らかの数値があった場合
					final char plc = (char)('①' + ( Integer.parseInt(lbl.substring( idx+1 ))-1 ) );
					editAttri.set( "placeholder" , String.valueOf( plc ) );
				}
			}

			// 理由 があれば、title に入れて popup させる
			final String riyu= rowData( row,noRiyu,"" );
			if( ! riyu.isEmpty() ) {
				editAttri.set( "title" , riyu );
			}

			config.setEditorAttributes( editAttri );				// class追加。元のｸﾗｽ属性に、登録方法(CDREC) と 判定結果(JUDG) を追加

			if( "3".equals( cdrec ) ) {								// 3:不要 は、writable を false にする。
				config.setWritable( "false" );
			}

			// 入力欄を連結するため、ﾙｰﾌﾟの最後にまとめて出す必要がある。
			// dbType が、TP_MENU の場合は、DBMENUの引数($2)に、ﾄｰｸﾝを渡す。(GG03を検索する)
			if( "TP_MENU".equals( dbType ) ) {
				config.setAddNoValue( true );						// プルダウンメニューは、空欄を選択できるようにしておく
				valBuf.append( new DBColumn( config ).getEditorValue( row,val+":" + token ) );	// 入力欄
			}
			else {
				valBuf.append( new DBColumn( config ).getEditorValue( row,val ) );				// 入力欄
			}

			final int colspan = rowData( row,noColspan,1 );
	//		final int rowspan = rowData( row,noRowspan,1 );

			// ﾗﾍﾞﾙ項目の表示
			if( "1".equals( cddisp ) || "4".equals( cddisp ) ) {	// CDDISP=1:ﾗﾍﾞﾙｶﾗﾑ,4:ﾗﾍﾞﾙのみ の場合、ﾗﾍﾞﾙ表示される
				if( lastTd ) {										// 前回、</td> を付けていないため
					valBufOut( row,out,valBuf,tani );				// out に、valBuf と tani と、警告異常範囲と終端 td を付ける。
					lastTd=false;
				}
	//			if( 1 == rowspan ) {
					out.append( "<td class='label'>" );
	//			}
	//			else {
	//				out.append( "<td class='label' rowspan=\"" ).append( rowspan ).append( "\">" );
	//			}
				out.append(lbl).append( "</td>" );
			}

			if( "1".equals( cddisp ) || "2".equals( cddisp ) ) {	// CDDISP=1:ﾗﾍﾞﾙｶﾗﾑ or 2:ｶﾗﾑ単発 の場合 td が必要。
				if( lastTd ) {										// 前回、</td> を付けていないため
					valBufOut( row,out,valBuf,tani );				// out に、valBuf と tani と、警告異常範囲と終端 td を付ける。
					lastTd=false;
				}
				out.append( "<td" );
				// colspan を反映させるのは、ﾌｨｰﾙﾄﾞ部のみ
				if( 1 != colspan ) {
					out.append( " colspan=\"" ).append( colspan*2+1 ).append( '"' );		// ラベルと本体があるので、colspan*2+1 となる
				}

	//			// rowspan を反映させるのは、ﾌｨｰﾙﾄﾞ部のみ
	//			if( 1 != rowspan ) {
	//				out.append( " rowspan=\"" ).append( rowspan ).append( '"' );
	//			}
				out.append( '>' );
			}

			lastTd = "3".equals( cddisp );							// CDDISP=3:ｶﾗﾑ連続 の場合 td は付けないが、後で必要になる。
			if( !lastTd ) {
				valBufOut( row,out,valBuf,tani );					// out に、valBuf と tani と、警告異常範囲と終端 td を付ける。
			}
		}

		if( lastTd ) {												// 前回、</td> を付けていないため
			valBufOut( lastNo-1,out,valBuf,tani );					// out に、valBuf と tani と、警告異常範囲と終端 td を付ける。
		}
		out.append(" </tr></tbody></table>").append( CR );
		if( !bkSelNm.isEmpty() ) { out.append( "</fieldset>" ); }
		out.append( getScrollBarEndDiv() );

		return out.toString();
	}

	/**
	 * 値出力の終端処理をまとめて行います。
	 *
	 * 全体出力の out に、値ﾌｨｰﾙﾄﾞの valBuf と、 tani を連結し、
	 * DBTableModelから、異常下限,警告下限,警告上限,異常上限を、div で固めます。
	 * valBuf は、初期化されます。
	 *
	 * @og.rev 7.3.1.1 (2021/02/25) USE_CSV 属性追加
	 *
	 * @param	row 行
	 * @param	out 全体出力のStringBuilder
	 * @param	valBuf 値ﾌｨｰﾙﾄﾞのStringBuilder(このﾒｿｯﾄﾞ内でｸﾘｱする)
	 * @param	tani 単位
	 */
	private void valBufOut( final int row,final StringBuilder out,final StringBuilder valBuf,final String tani ) {
		out.append( valBuf ).append( tani ).append( "<div class='small' title='上段:警告&#10;下段:異常'>" );

		final String wmin = rowData( row,noWmin,"" );		// 警告下限(W_MIN)
		final String wmax = rowData( row,noWmax,"" );		// 警告上限(W_MAX)
		if( !wmin.isEmpty() || !wmax.isEmpty() ) {
			final String max = "%" + Math.max( wmin.length() , wmax.length() ) + "s";	// 最大桁数を求める。
			out.append( String.format( max + "～" + max ,wmin,wmax ) );
		}
		out.append( "</br>" );

		final String emin = rowData( row,noEmin,"" );		// 異常下限(E_MIN)
		final String emax = rowData( row,noEmax,"" );		// 異常上限(E_MAX)
		if( !emin.isEmpty() || !emax.isEmpty() ) {
			final String max = "%" + Math.max( emin.length() , emax.length() ) + "s";	// 最大桁数を求める。
			out.append( String.format( max + " ～ " + max ,emin,emax ) );				// 異常の範囲の方が広いので、間を少し開けている。
		}
		out.append( "</div></td>" );

		valBuf.setLength(0);
	}

	/**
	 * DBTableModelから、指定行-列のデータを取得します。
	 *
	 * 列番号がマイナスの場合は、ｶﾗﾑが存在していないため、初期値を返します。
	 *
	 * @og.rev 7.3.1.1 (2021/02/25) USE_CSV 属性追加
	 *
	 * @param	row 行
	 * @param	col 列
	 * @param	defVal 初期値(文字列)
	 *
	 * @return 指定行-列のデータ(文字列)
	 */
	private String rowData( final int row , final int col , final String defVal ) {
		String rtn = defVal ;
		if( col >= 0 ) {
			final String str = table.getValue( row,col );
			if( str != null && !str.isEmpty()) {
				rtn = str ;
			}
		}
		return rtn ;
	}

	/**
	 * DBTableModelから、指定行-列のデータを取得し、前後に文字列を追加します。
	 *
	 * 列番号がマイナスの場合は、ｶﾗﾑが存在していないため、長さゼロの文字列を返します。
	 * 取得したデータが、長さゼロの文字列でない場合は、前後に指定の文字列を追加して返します。
	 *
	 * @og.rev 7.3.1.1 (2021/02/25) USE_CSV 属性追加
	 *
	 * @param	row 行
	 * @param	col 列
	 * @param	prefix 前に付ける文字列
	 * @param	suffix 後ろに付ける文字列
	 *
	 * @return 指定行-列のデータ(文字列)
	 */
	private String rowData( final int row , final int col , final String prefix , final String suffix ) {
		String rtn = "" ;
		if( col >= 0 ) {
			final String str = table.getValue( row,col );
			if( str != null && !str.isEmpty()) {
				rtn = prefix + str + suffix ;
			}
		}
		return rtn ;
	}

	/**
	 * DBTableModelから、指定行-列のデータを取得します。
	 *
	 * 列番号がマイナスの場合は、ｶﾗﾑが存在していないため、初期値を返します。
	 *
	 * @og.rev 7.3.1.1 (2021/02/25) USE_CSV 属性追加
	 *
	 * @param	row 行
	 * @param	col 列
	 * @param	defVal 初期値(数値)
	 *
	 * @return 指定行-列のデータ(数値)
	 */
	private int rowData( final int row , final int col , final int defVal ) {
		final String val = col < 0 ? null : table.getValue( row,col ) ;

		try {
			return val == null || val.isEmpty() ? defVal : Integer.parseInt( val );
		}
		catch( final NumberFormatException ex ) {
			final String errMsg = "数値項目に数値以外の文字が含まれています。行=" + row + " , 列=" + col + " , 値=" + val ;
			throw new HybsSystemException( errMsg,ex );
		}
	}

	/**
	 * DBTableModel から テーブルのヘッダータグ文字列を作成して返します。
	 *
	 * @og.rev 3.5.2.0 (2003/10/20) ヘッダーそのもののキャッシュはしない。
	 *
	 * @return	テーブルのヘッダータグ文字列
	 * @og.rtnNotNull
	 */
	protected String getHeader() {
		// ヘッダーは不要なので、オーバーライドしておく。
		return "" ;
	}
}
